package backend;

import compiler.*;
import intermediate.frame.*;
import intermediate.operators.*;
import intermediate.*;
import java.util.*;
import java.io.*;

public class InterpreterBackend implements Backend {

    private compiler.Compiler compiler;
    private int PC;
    private List currentCode;
    private int currentScope;
    private int currentLevel;
    private HashMap temps;
    private HashMap labelTable;
    private ArrayList stack;
    private List codelist;
    private FrameContainer fc;
    private int bp;
    private int fp;

    public InterpreterBackend(compiler.Compiler compiler) {
        this.compiler = compiler;
    }

    public compiler.Compiler getCompiler() {
        return compiler;
    }

    /**
     * Gibt ein Dummyframe zurck, bentigt fr einige Berechnungen im Frontend (z.B. bei wordsize)
     */
    public Frame getFrame() {
        return new InterpreterFrame();
    }

    /**
     * Der eigentliche Einstiegspunkt und die Hauptschleife des Interpreters
     */
    public void transformToExecutable(List codelist, FrameContainer fc) {
        this.codelist = codelist;
        this.fc = fc;
        //Reset des Interpreters
        currentScope = 0;
        currentLevel = 0;
        stack = new ArrayList();
        //stack.push(fc.getFrame(currentScope));
        stack.addAll(fc.getFrame(currentScope).getFields());
        fp = 0;
        bp = stack.size();
        PC = 0;
        currentCode = (List)codelist.get(currentScope);
        temps = new HashMap();
        labelTable = new HashMap();
        setupLabelTable(currentCode);
        //solange der PC noch nicht am Ende eines Blockes ist oder solange noch nicht aus allen
        //Blcken zurckgesprungen wurde
        while (PC < currentCode.size() || currentScope != 0) {
            //Wenn der PC noch nicht am Ende ist, fhre nchsten Befehl aus
            if (PC < currentCode.size()) {
                Object nextInstr = currentCode.get(PC);
                Statement stm = (Statement)nextInstr;
                executeStatement(stm);
                PC++;
            }
            //Wenn der PC am Ende ist und noch Blcke aktiv sind, fhre ein Return aus der
            //Prozedur aus
            if (currentScope != 0 && PC >= currentCode.size()) {
                prepareReturn();
            }
            //ansonsten ist das Programm abgearbeitet
        }
    }

    public void prepareCall(CALL call) {
        try {
            //Nummer der aufzurufenden Prozedur
            int callNumber = processExpression(call.getProc());
            InterpreterFrame iframe = fc.getFrame(callNumber).createClone();
            //neues Frame auf Stack legen
            stack.addAll(iframe.getFields());
            //dynamischer Link
            stack.set(bp + 1, new Integer(fp));
            //PC
            stack.set(bp + 2, new Integer(PC));
            //aktueller Codeblock
            stack.set(bp + 3, new Integer(currentScope));
            //System.out.println("currentLevel: " + currentLevel);
            //System.out.println("newLevel: " + iframe.level());
            //Neuer statischer Link: Wenn a b aufruft, folge level(a) - level(b) + 1
            //statischen Links von a aus, das Ziel ist der statische Link von b
            int newStaticLink = fp; //im einfachsten Fall zeigt der static link auf das aufrufende Frame
            for (int i = 0; i < currentLevel - iframe.level() + 1; i++) {
                newStaticLink = ((Integer)stack.get(newStaticLink + 0)).intValue();
            }
            stack.set(bp + 0, new Integer(newStaticLink));
            fp = bp;
            bp = bp + iframe.getNumberOfFields() + iframe.getMinOffset();
            currentCode = (List)codelist.get(callNumber);
            currentScope = callNumber;
            currentLevel = iframe.level();
            PC = -1;
            setupLabelTable(currentCode);
            /*System.out.println("end of prepare for " + callNumber + ":");
            System.out.println("fp: " + fp);
            System.out.println("bp: " + bp);
            System.out.println(stackDump());*/

        }
        catch (OutOfMemoryError oome) {
            System.err.println("Runtime Error: Out of memory");
            System.exit( -3);
        }
    }

    public void prepareReturn() {
        bp = fp;
        fp = ((Integer)stack.get(bp + 1)).intValue();
        PC = ((Integer)stack.get(bp + 2)).intValue() + 1;
        currentScope = ((Integer)stack.get(bp + 3)).intValue();
        currentCode = (List)codelist.get(currentScope);
        //Berechne den aktuellen Level nach Verlassen dieses Frames. Zhle dazu die Anzahl der
        //statischen Links, die bis zu Startadresse (0) fhren.
        int ptr = bp;
        currentLevel = 0;
        while (ptr != 0) {
            ptr = ((Integer)stack.get(ptr + 0)).intValue();
            currentLevel++;
        }
        //System.out.println("level: " + currentLevel);
        removeFromStack(stack.size() - bp);
        /*System.out.println("leaving");
        System.out.println("fp: " + fp);
        System.out.println("bp: " + bp);
        System.out.println(stackDump());*/
    }

    private void removeFromStack(int amountToRemove) {
        for (int i = 0; i < amountToRemove; i++) {
            stack.remove(stack.size() - 1);
        }
    }

    /**
     * Trgt alle Label und deren Position in eine Tabelle ein und entfernt die Label aus dem
     * auszufhrenden Code
     */
    private void setupLabelTable(List currentCode) {
        int i = 0;
        for (Iterator it = currentCode.iterator(); it.hasNext(); ) {

            Object o = it.next();
            if (o instanceof LABEL) {
                LABEL label = (LABEL)o;
                //i - 1, weil in der Hauptschleife der PC anschlieend immer 1 hochgezhlt wird,
                //darum setzen wir das Label auf 1 weniger. Nach der Erhhung steht der PC an der
                //richtigen Stelle
                labelTable.put(label.getLabel(), new Integer(i - 1));
                it.remove();
            } else {
                i++;
            }
        }

    }

    private void executeStatement(Statement s) {
        //MOVE
        if (s instanceof MOVE) {
            MOVE move = (MOVE)s;
            int result = processExpression(move.getSrc());
            storeResult(move.getDest(), result);
        }
        //JUMP
        if (s instanceof JUMP) {
            JUMP jump = (JUMP)s;
            PC = ((Integer)labelTable.get(jump.getLabel())).intValue();
        }
        //CJUMP
        if (s instanceof CJUMP) {
            CJUMP cjump = (CJUMP)s;
            boolean guardexpr = false;
            int leftop = processExpression(cjump.getLeftOp());
            int rightop = processExpression(cjump.getRightOp());
            switch (cjump.getRelopInt()) {
                case CJUMP.EQ:
                    guardexpr = (leftop == rightop);
                    break;
                case CJUMP.NE:
                    guardexpr = (leftop != rightop);
                    break;
                case CJUMP.LT:
                    guardexpr = (leftop < rightop);
                    break;
                case CJUMP.GT:
                    guardexpr = (leftop > rightop);
                    break;
                case CJUMP.LE:
                    guardexpr = (leftop <= rightop);
                    break;
                case CJUMP.GE:
                    guardexpr = (leftop > rightop);
                    break;
            }
            if (guardexpr) {
                PC = ((Integer)labelTable.get(cjump.getTrueLabel())).intValue();
            } else {
                PC = ((Integer)labelTable.get(cjump.getFalseLabel())).intValue();
            }
        }
        //OUTPUT
        if (s instanceof OUTPUT) {
            OUTPUT output = (OUTPUT)s;
            int value = processExpression(output.getLoc());
            System.out.println(value);
        }
        //INPUT
        if (s instanceof INPUT) {
            System.out.print("Input: ");
            INPUT input = (INPUT)s;
            MEM address = input.getAddress();
            BufferedReader br;
            br = new BufferedReader(new InputStreamReader(System.in));
            int result = 0;
            try {
                result = Integer.parseInt(br.readLine());
            }
            catch (IOException ex) {
            }
            catch (NumberFormatException ex) {
                System.err.println("Runtime Error: Input must be a number between -2147483648 and 2147483647");
                System.exit( -1);
            }
            storeResult(address, result);
        }
        //Alle anderen Flle, die keine Spezialbehandlung brauchen
        //z.B. EXP
        processExpression((Expression)s.getFirstChild());

    }

    /**
     * Wertet Expressions aus und liefert deren Ergebnis zurck. Da bereits linearisiert wurde,
     * sind die Expressions nicht mehr geschachtelt und mssen nicht mehr rekursiv abgearbeitet
     * werden.
     */
    public int processExpression(Expression e) {
        //Konstanten (Zahlen im Quelltext)
        if (e instanceof CONST) {
            CONST con = (CONST)e;
            return con.getValue();
        }
        //Temporaries (Register)
        if (e instanceof TEMP) {
            TEMP t = (TEMP)e;
            //Der FramePointer.
            if (t.getTemp().toString().equals("Temp fp")) {
                return fp;
            }
            //Normale Temps
            return ((Integer)temps.get(t.getTemp())).intValue();
        }
        if (e instanceof MEM) {
            MEM m = (MEM)e;
            //von welcher Adresse soll geladen werden?
            int location = processExpression(m.getLoc());
            int value = loadFromStack(location);
            //currentLookup = null;
            return value;
        }
        if (e instanceof BINOP) {
            BINOP binop = (BINOP)e;
            //Operatoren ermitteln
            int op1 = processExpression(binop.getSrc1());
            int op2 = processExpression(binop.getSrc2());
            switch (binop.getOperatorInt()) {
                case BINOP.PLUS:
                    return op1 + op2;
                case BINOP.MINUS:
                    return op1 - op2;
                case BINOP.MUL:
                    return op1 * op2;
                case BINOP.DIV:
                    if (op2 == 0) {
                        System.err.println("Runtime Error: Division by zero");
                        System.exit( -2);
                    }
                    return op1 / op2;
                case BINOP.AND:
                    return op1 & op2;
                case BINOP.OR:
                    return op1 | op2;
                case BINOP.XOR:
                    return op1 ^ op2;
                case BINOP.LSHIFT:
                    return op1 << op2;
                case BINOP.RSHIFT:
                    return op1 >> op2;
            }

        }
        if (e instanceof CALL) {
            CALL call = (CALL)e;
            prepareCall(call);
        }
        return -1;
    }

    /**
     * Speichert den Wert result ab. Dabei wird unterschieden, ob dest ein Register (TEMP) oder
     * Hauptspeicher (MEM) ist.
     */
    public void storeResult(Expression dest, int result) {
        if (dest instanceof TEMP) {
            Temp t = ((TEMP)dest).getTemp();
            temps.put(t, new Integer(result));
        }
        if (dest instanceof MEM) {
            int location = processExpression(((MEM)dest).getLoc());
            writeOnStack(location, result);
        }
    }

    /**
     * Ldt den Wert von Adresse address aus dem simulierten Hauptspeicher (Stack)
     */
    private int loadFromStack(int address) {
        return ((Integer)stack.get(address)).intValue();
    }

    /**
     * Speichert den Wert value an Adresse address in den simulierten Hauptspeicher (Stack)
     */
    private void writeOnStack(int address, int value) {
        stack.set(address, new Integer(value));
    }

    /**
     * Hilfsmethode zum Debuggen, der Inhalt des simulierten Hauptspeichers wird ausgegeben.
     */
    private String stackDump() {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < stack.size(); i++) {
            sb.append(i + ": " + stack.get(i));
            sb.append("\n");
        }
        return sb.toString();
    }
}
